from ewokscore import TaskWithProgress

from ImageD11 import frelon_peaksearch

from ..models import (
    SegmenterConfig,
    SegmenterFolderConfig,
    SegmenterCorrectionFiles,
)
from ..utils import TqdmProgressCallback, extract_sample_info, get_monitor_scale_factor
from pathlib import Path
import h5py
from ..nexus import save_nexus_process, group_exists


class SegmentScan(
    TaskWithProgress,
    input_names=["folder_config", "segmenter_algo_params", "correction_files"],
    optional_input_names=["monitor_name", "overwrite"],
    output_names=[
        "sample_folder_info",
        "segmented_peaks_url",
    ],
):
    """
    This task segments an entire scan folder,
    merges the peaks, and produces a 3D column file.
    The resulting 3D column peak file is saved.

    Outputs:

    - `segmented_3d_peaks_file`: A segmented 3d peaks for the given scan folder is saved in this path.
    - `sample_folder_info`: A Config information about raw scan sample
    - `segmented_peaks_url`: A Nexus file data url path to `segmented_3d_peaks` data
    """

    def run(self):

        seg_folder_config = SegmenterFolderConfig(**self.inputs.folder_config)

        detector = seg_folder_config.detector
        omega_motor = seg_folder_config.omega_motor
        scan_folder = seg_folder_config.scan_folder

        _, sample_name, dset_name, scan_number = extract_sample_info(
            path_str=scan_folder
        )

        masterfile_path = Path(scan_folder).parent / f"{sample_name}_{dset_name}.h5"
        if not masterfile_path.exists():
            raise FileNotFoundError(
                """ 'scan_folder' field in 'seg_folder_config' does not have master file
                        in its parent folder. From its parent folder, we will have its masterfile path.
                    This masterfile path required to compute Omega (rotation) values.
                """
            )
        analyse_folder = seg_folder_config.analyse_folder
        output_folder = (
            Path(analyse_folder) / sample_name / f"{sample_name}_{dset_name}"
        )
        output_folder.mkdir(parents=True, exist_ok=True)
        nexus_file_path = output_folder / f"{sample_name}_{dset_name}.h5"
        seg_data_group_path = f"{str(scan_number)}.1/segmented_3d_peaks"
        overwrite = self.get_input_value("overwrite", False)
        if not overwrite and group_exists(
            filename=nexus_file_path, data_group_path=seg_data_group_path
        ):
            raise ValueError(
                f"""Data group '{seg_data_group_path}' already exists in {nexus_file_path},
                Set `overwrite` to True if you wish to overwrite the existing data group.
                """
            )

        with h5py.File(str(masterfile_path), "r") as hin:
            omega_angles = hin[str(scan_number) + ".1"]["measurement"].get(
                omega_motor, None
            )
            omega_array = omega_angles[()]

        segmenter_cfg = SegmenterConfig(**self.inputs.segmenter_algo_params)
        correction_files_config = SegmenterCorrectionFiles(
            **self.inputs.correction_files
        )
        segmenter_settings = {
            "bgfile": correction_files_config.bg_file,
            "maskfile": correction_files_config.mask_file,
            "darkfile": correction_files_config.dark_file,
            "flatfile": correction_files_config.flat_file,
            "threshold": segmenter_cfg.threshold,
            "smoothsigma": segmenter_cfg.smooth_sigma,
            "bgc": segmenter_cfg.bgc,
            "minpx": segmenter_cfg.min_px,
            "m_offset_thresh": segmenter_cfg.offset_threshold,
            "m_ratio_thresh": segmenter_cfg.ratio_threshold,
        }

        monitor_name = self.get_input_value("monitor_name", None)
        scale_factor = (
            None
            if monitor_name is None
            else get_monitor_scale_factor(masterfile_path, scan_number, monitor_name)
        )

        all_frames_2d_peaks_list = frelon_peaksearch.segment_master_file(
            str(masterfile_path),
            str(scan_number) + ".1" + "/measurement/" + detector,
            omega_array,
            segmenter_settings,
            scale_factor=scale_factor,
            tqdm_class=TqdmProgressCallback,
            TaskInstance=self,
        )
        peaks_2d_dict, num_peaks = frelon_peaksearch.peaks_list_to_dict(
            all_frames_2d_peaks_list
        )
        # 3d merge from 2d peaks dict
        peak_3d_dict = frelon_peaksearch.do3dmerge(
            peaks_2d_dict, num_peaks, omega_array
        )

        self.outputs.sample_folder_info = {
            "omega_motor": omega_motor,
            "scan_folder": scan_folder,
        }

        nxprocess_url = save_nexus_process(
            filename=nexus_file_path,
            entry_name=f"{str(scan_number)}.1",
            process_name="segmented_3d_peaks",
            peaks_data=peak_3d_dict,
            config_settings={
                "FolderFileSettings": self.inputs.folder_config,
                "Segmenter_settings": segmenter_settings,
            },
            pks_axes=("f_raw", "s_raw"),
            signal_name="Number_of_pixels",
            scale="log",
            overwrite=overwrite,
        )
        self.outputs.segmented_peaks_url = nxprocess_url
